@@ -108,13 +108,11 @@ char *strchr(), *strrchr();
 #endif
 
 
-#ifdef HAVE_PTY_H
+#ifdef HAVE_PTY_H /* glibc (even if BSD) */
 #  include <pty.h>
-#endif
-#ifdef HAVE_LIBUTIL_H
+#elif HAVE_LIBUTIL_H /* BSD, non-glibc */
 #  include <libutil.h>
-#endif
-#ifdef HAVE_UTIL_H
+#elif HAVE_UTIL_H /* BSD, other varriants */
 #  include <util.h>
 #endif
 
@@ -129,14 +127,23 @@ char *strchr(), *strrchr();
 
 
 #ifdef  HAVE_SNPRINTF           /* don't rely on the compiler understanding variadic macros */
-# define snprintf0(buf,bufsize,format)                  snprintf(buf,bufsize,format)
-# define snprintf1(buf,bufsize,format,arg1)             snprintf(buf,bufsize,format,arg1)
-# define snprintf2(buf,bufsize,format,arg1,arg2)        snprintf(buf,bufsize,format,arg1,arg2)
+# define snprintf0(buf,bufsize,format)                           snprintf(buf,bufsize,format)
+# define snprintf1(buf,bufsize,format,arg1)                      snprintf(buf,bufsize,format,arg1)
+# define snprintf2(buf,bufsize,format,arg1,arg2)                 snprintf(buf,bufsize,format,arg1,arg2)
+# define snprintf3(buf,bufsize,format,arg1,arg2,arg3)            snprintf(buf,bufsize,format,arg1,arg2,arg3)
+# define snprintf4(buf,bufsize,format,arg1,arg2,arg3,arg4)       snprintf(buf,bufsize,format,arg1,arg2,arg3,arg4)
+# define snprintf5(buf,bufsize,format,arg1,arg2,arg3,arg4,arg5)  snprintf(buf,bufsize,format,arg1,arg2,arg3,arg4,arg5)
+# define snprintf6(buf,bufsize,format,arg1,arg2,arg3,arg4,arg5,arg6)  snprintf(buf,bufsize,format,arg1,arg2,arg3,arg4,arg5,arg6)
+
 #else
-# define snprintf0(buf,bufsize,format)                  sprintf(buf,format)
-# define snprintf1(buf,bufsize,format,arg1)             sprintf(buf,format,arg1)
-# define snprintf2(buf,bufsize,format,arg1,arg2)        sprintf(buf,format,arg1,arg2)
-# define vsnprintf(buf,bufsize,format,ap)               vsprintf(buf,format,ap)
+# define snprintf0(buf,bufsize,format)                           sprintf(buf,format)
+# define snprintf1(buf,bufsize,format,arg1)                      sprintf(buf,format,arg1)
+# define snprintf2(buf,bufsize,format,arg1,arg2)                 sprintf(buf,format,arg1,arg2)
+# define snprintf3(buf,bufsize,format,arg1,arg2,arg3)            sprintf(buf,format,arg1,arg2,arg3)
+# define snprintf4(buf,bufsize,format,arg1,arg2,arg3,arg4)       sprintf(buf,format,arg1,arg2,arg3,arg4)
+# define snprintf5(buf,bufsize,format,arg1,arg2,arg3,arg4,arg5)  sprintf(buf,format,arg1,arg2,arg3,arg4,arg5)
+# define snprintf6(buf,bufsize,format,arg1,arg2,arg3,arg4,arg5,arg6)  sprintf(buf,format,arg1,arg2,arg3,arg4,arg5,arg6)
+# define vsnprintf(buf,bufsize,format,ap)                        vsprintf(buf,format,ap)
 #endif
 
 
@@ -181,13 +188,14 @@ extern int _rl_horizontal_scroll_mode;
 
 /* in main.c: */
 extern int master_pty_fd;
-extern int slave_pty_fd;
+extern int slave_pty_sensing_fd;
 extern FILE *debug_fp;
 extern char *program_name, *command_name;
 extern int always_readline;
 extern int complete_filenames;
 extern pid_t command_pid;
 extern char *command_line;
+extern char *extra_char_after_completion;
 extern int i_am_child;
 extern int nowarn;
 extern int debug;
@@ -196,6 +204,7 @@ extern int one_shot_rlwrap;
 extern char *substitute_prompt;
 extern char *history_format;
 extern char *forget_regexp;
+extern char *multi_line_tmpfile_ext;
 extern char *prompt_regexp;
 extern int renice;
 extern int ignore_queued_input;
@@ -215,6 +224,8 @@ extern int we_just_got_a_signal_or_EOF;
 extern int remember_for_completion;
 extern int commands_children_not_wrapped; 
 extern int accepted_lines;
+extern char *filter_command;
+extern int polling;
 
 void cleanup_rlwrap_and_exit(int status);
 void put_in_output_queue(char *stuff);
@@ -253,7 +264,7 @@ extern int leave_prompt_alone;
 
 
 /* in signals.c */
-extern int command_is_dead;
+extern volatile int command_is_dead;
 extern int commands_exit_status;
 extern int filter_is_dead;
 extern int filters_exit_status;
@@ -295,9 +306,18 @@ void  write_patiently2(int fd, const void *buffer, int count, int uninterruptibl
 void  mysetenv(const char *name, const char *value);
 void  set_ulimit(int resource, long value);
 void  usage(int status);
-void  mywarn(const char *message, ...);
-void  myerror(const char *message, ...);
-void *mymalloc(size_t size);
+int   open_unique_tempfile(const char *suffix, char **tmpfile_name);
+
+
+/* flags to use for the error_flags argument to myerror */
+#define FATAL     2
+#define WARNING   0
+#define USE_ERRNO 1
+#define NOERRNO   0
+
+
+void  myerror(int error_flags, const char *message, ...);
+void  *mymalloc(size_t size);
 void  myfree(void *ptr);
 void  mysetsid(void);
 void  close_open_files_without_writing_buffers(void);
@@ -327,6 +347,7 @@ char *add3strings(const char *str1, const char *str2, const char *str3);
 #define add2strings(a,b)  add3strings(a,b,"")
 char *mystrtok(const char *s, const char *delim);
 char **split_with(const char *string, const char *delim);
+char *unsplit_with(int n, char ** strings, const char *delim);
 char **split_on_single_char(const char *string, char c);
 int scan_metacharacters(const char* string, const char *metacharacters);
 char **list4 (char *el0, char *el1, char *el2, char *el3);
@@ -360,6 +381,7 @@ struct termios *get_pterm_slave(void);
 void mirror_slaves_echo_mode(void);
 void completely_mirror_slaves_terminal_settings(void);
 void completely_mirror_slaves_output_settings(void);
+void completely_mirror_slaves_special_characters(void);
 void write_EOF_to_master_pty(void);
 void write_EOL_to_master_pty(char *);
 int dont_wrap_command_waits(void);
@@ -393,6 +415,7 @@ void clear_line(void);
 void clear_the_screen(void);
 void curs_up(void);
 void curs_down(void);
+void curs_left(void);
 void test_terminal(void);
 int my_putchar(TPUTS_PUTC_ARGTYPE c);
 void my_putstr(const char *string);
@@ -402,7 +425,7 @@ extern int terminal_settings_saved;
 extern struct winsize winsize;
 extern char *term_name;
 extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos,
-  *term_cursor_up, *term_cursor_down;
+  *term_cursor_up, *term_cursor_down, *term_cursor_left, *term_rmcup, *term_rmkx;
 extern int newline_came_last;
 
 /* in filter.c */
@@ -518,7 +541,7 @@ void filter_test(void);
 
 #  define ERRMSG(b)              (b && (errno != 0) ? add3strings("(", strerror(errno), ")") : "" )
 
-#  define SHOWCURSOR(c)          if (debug & DEBUG_READLINE) {my_putchar(c); mymicrosleep(800); backspace(1);} /* (may work incorrectly at last column!)*/
+#  define SHOWCURSOR(c)          if (debug & DEBUG_READLINE) {my_putchar(c); mymicrosleep(1200); curs_left();} /* (may work incorrectly at last column!)*/
 
 #  define DEBUG_RANDOM_SLEEP        if (debug & DEBUG_RACES) {int sleeptime=rand()&31; DPRINTF1(DEBUG_RACES,"sleeping for %d msecs", sleeptime); mymicrosleep(sleeptime);}
 
