@@ -20,7 +20,7 @@
 
 #include "rlwrap.h"
 
-int command_is_dead = FALSE;
+volatile int command_is_dead = FALSE;
 int commands_exit_status = 0;
 int filter_is_dead = FALSE;
 int filters_exit_status = 0;
@@ -75,7 +75,7 @@ mysignal(int sig, sighandler_type handler) {
   if (signal(sig, handler) == SIG_ERR)
 # endif
     if(handler != SIG_DFL) /* allow e.g. KILL to be set to its default */ 
-      myerror("Error setting handler for signal %d (%s)", sig, signal_name(sig));   
+      myerror(FATAL|USE_ERRNO, "Failed setting handler for signal %d (%s)", sig, signal_name(sig));   
 }
 
 
@@ -85,10 +85,10 @@ install_signal_handlers()
   int i;
   mysignal(SIGCHLD, &child_died);
   mysignal(SIGTSTP, &handle_sigTSTP);
-#ifndef DEBUG
+#ifndef DEBUG          /* we want core dumps when debugging, no polite excuses! */
   for (i = 1; i<MAX_SIG; i++)
     if (signals_program_error(i))
-      mysignal(i, &handle_program_error_signal);        /* we want core dumps when debugging, no polite excuses! */
+      mysignal(i, &handle_program_error_signal); /* make polite excuse */
 #endif
   mysignal(SIGALRM, &handle_sigALRM);
   for (i = 1;  signals_to_be_passed_on[i]; i++) {
@@ -186,7 +186,7 @@ handle_sigTSTP(int signo)
   zero_select_timeout();
   /* Hand the SIGTSTP down to command and its process group */
   if (command_pid && (error = kill(-command_pid, SIGTSTP))) {
-    myerror("Failed to deliver SIGTSTP");
+    myerror(FATAL|USE_ERRNO, "Failed to deliver SIGTSTP");
   }
 
   if (within_line_edit)
@@ -210,7 +210,7 @@ handle_sigTSTP(int signo)
 
   #ifdef __QNX__
   if (command_pid && (error = kill(-command_pid, SIGCONT))) {
-    myerror("Failed to deliver SIGCONT");
+    myerror(FATAL|USE_ERRNO, "Failed to deliver SIGCONT");
   }
   #endif
   
@@ -291,8 +291,9 @@ adapt_tty_winsize(int from_fd, int to_fd)
   ret = ioctl(from_fd, TIOCGWINSZ, &winsize);
   DPRINTF1(DEBUG_SIGNALS, "ioctl (..., TIOCGWINSZ) = %d", ret);
   if (winsize.ws_col != old_winsize.ws_col || winsize.ws_row != old_winsize.ws_row) { 
-    if (always_readline &&!dont_wrap_command_waits())                       /* if --always_readline option is set, the client will probably spew a */
-      deferred_adapt_commands_window_size = TRUE;   /* volley of control chars at us when its terminal is resized. Hence we don't do it now */
+    DPRINTF4(DEBUG_SIGNALS, "ws.col: %d -> %d, ws.row: %d -> %d", old_winsize.ws_col, winsize.ws_col, old_winsize.ws_row, winsize.ws_row);
+    if (always_readline &&!dont_wrap_command_waits())  /* if --always_readline option is set, the client will probably spew a */
+      deferred_adapt_commands_window_size = TRUE;      /* volley of control chars at us when its terminal is resized. Hence we don't do it now */
     else {  
       ret = ioctl(to_fd, TIOCSWINSZ, &winsize); 
       DPRINTF1(DEBUG_SIGNALS, "ioctl (..., TIOCSWINSZ) = %d", ret);
@@ -328,6 +329,7 @@ void wipe_textarea(struct winsize *old_winsize)
     promptlength = colourless_strlen((saved_rl_state.cooked_prompt ? saved_rl_state.cooked_prompt:  saved_rl_state.raw_prompt), NULL, old_winsize -> ws_col); 
     linelength = (within_line_edit ? strlen(rl_line_buffer) : 0) + promptlength;
     point = (within_line_edit ? rl_point : 0) + promptlength;
+    assert(old_winsize -> ws_col > 0);
     lineheight = (linelength == 0 ? 0 : (1 + (max(point, (linelength - 1)) / old_winsize -> ws_col)));
     if (lineheight > 1 && term_cursor_up != NULL && term_cursor_down != NULL) {
       /* i.e. if we have multiple rows displayed, and we can clean them up first */
@@ -364,14 +366,14 @@ child_died(int unused)
     filter_is_dead = TRUE;
     filter_pid = 0;
   } else  {
-    DPRINTF0(DEBUG_ALL, "Whoa, got a SIGCHLD without one of the children actually dying....!");
+    DPRINTF0(DEBUG_ALL, "Whoa, got a SIGCHLD, but not from slave command or filter! I must have children I don't know about (blush...)!");
     /* ignore */
   }     
 
   errno = saved_errno;  
-  return;                       /* allow remaining output from child to be processed in main loop */
-  /* (so that we will see childs good-bye talk)                     */
-  /* this will then clean up and terminate                          */
+  return;   /* allow remaining output from child to be processed in main loop */
+            /* (so that we will see childs good-bye talk)                     */
+            /* this will then clean up and terminate                          */
 
 }
 
@@ -415,12 +417,11 @@ void suicide_by(int signal, int status) {
      the shell may tell the user that rlwrap itself has failed. Make clear that
      it didn't. @@@ We could also try argv[0] = command_name just before dying ? */
   
-  if (signals_program_error(signal)) { 
-    errno = 0;  
-    mywarn("%s killed by %s%s.\n%s has not crashed, but for transparency,\n"
+  if (signals_program_error(signal)) {  
+    myerror(WARNING|NOERRNO, "%s crashed, killed by %s%s.\n%s itself has not crashed, but for transparency,\n"
            "it will now kill itself %swith the same signal\n",
            command_name, signal_name(signal), (coredump(status) ? " (core dumped)" : ""),
-           program_name, (coredump(status) ? "" : "(without dumping core)") );
+           program_name, (coredump(status) ? "" : "(without dumping core) ") );
   }
   uninstall_signal_handlers();
   unblock_all_signals();
